##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import logging

from .test_utils import QasmFile, get_phir_json

logger = logging.getLogger(__name__)


def test_parallelization() -> None:
    """Make sure the parallelization is happening properly for the test circuit."""
    phir = get_phir_json(QasmFile.parallelization_test, rebase=True)

    # Make sure The parallel RZ and R1XY gates have the correct arguments
    parallel_rz1 = phir["ops"][3]
    assert parallel_rz1["qop"] == "RZ"
    qubits = [["q", 0], ["q", 1], ["q", 2], ["q", 3]]
    for qubit in qubits:
        assert qubit in parallel_rz1["args"]
    parallel_phasedx = phir["ops"][5]
    assert parallel_phasedx["qop"] == "R1XY"
    for qubit in qubits:
        assert qubit in parallel_phasedx["args"]

    # Make sure the parallel block is properly formatted
    block = phir["ops"][7]
    assert block["block"] == "qparallel"
    assert len(block["ops"]) == 2
    qop0 = block["ops"][0]
    qop1 = block["ops"][1]
    assert qop0["qop"] == qop1["qop"] == "RZZ"

    # Make sure the ops within the parallel block have the correct arguments
    assert len(qop0["args"][0]) == len(qop1["args"][0]) == 2
    q01_fst = (["q", 0] in qop0["args"][0]) and (["q", 1] in qop0["args"][0])
    q01_snd = (["q", 0] in qop1["args"][0]) and (["q", 1] in qop1["args"][0])
    q23_fst = (["q", 2] in qop0["args"][0]) and (["q", 3] in qop0["args"][0])
    q23_snd = (["q", 2] in qop1["args"][0]) and (["q", 3] in qop1["args"][0])
    assert (q01_fst and q23_snd) != (q23_fst and q01_snd)

    # Make sure the measure op is properly formatted
    measure = phir["ops"][9]
    measure_args = measure["args"]
    measure_returns = measure["returns"]
    assert len(measure_args) == len(measure_returns) == 4
    assert measure_args.index(["q", 0]) == measure_returns.index(["c", 0])
    assert measure_args.index(["q", 1]) == measure_returns.index(["c", 1])
    assert measure_args.index(["q", 2]) == measure_returns.index(["c", 2])
    assert measure_args.index(["q", 3]) == measure_returns.index(["c", 3])


def test_parallel_subcommand_relative_ordering() -> None:
    """Make sure the proper relative ordering of sub-commands is preserved."""
    phir = get_phir_json(QasmFile.rxrz, rebase=True)
    # make sure it is ordered like the qasm file
    ops = phir["ops"]
    frst_sc = ops[3]
    scnd_sc = ops[5]
    thrd_sc = ops[7]
    frth_sc = ops[9]
    assert frst_sc["qop"] == "RZ"
    assert frst_sc["angles"] == [[0.5], "pi"]
    assert scnd_sc["qop"] == "R1XY"
    assert scnd_sc["angles"] == [[3.5, 0.0], "pi"]
    assert thrd_sc["qop"] == "R1XY"
    assert thrd_sc["angles"] == [[0.5, 0.0], "pi"]
    assert frth_sc["qop"] == "RZ"
    assert frth_sc["angles"] == [[3.5], "pi"]



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from typing import TYPE_CHECKING, TypeAlias

if TYPE_CHECKING:
    from .shard import Shard, ShardLayer

Layer: TypeAlias = list[list[int]]


def parse_shards_naive(
    shards: set["Shard"],
) -> tuple[list[Layer], list["ShardLayer"]]:
    """Parse a set of shards and return a circuit representation for placement."""
    layers: list[Layer] = []
    shards_in_layer: list[ShardLayer] = []
    scheduled: set[int] = set()
    num_shards: int = len(shards)

    while len(scheduled) < num_shards:
        layer: Layer = []

        # Iterate the shards, looking for shards whose dependencies have been
        # satisfied, or initially, shards with no dependencies
        to_schedule: ShardLayer = [
            s for s in shards if s.depends_upon.issubset(scheduled)
        ]
        shards_in_layer.append(to_schedule)
        shards.difference_update(to_schedule)

        for shard in to_schedule:
            op: list[int] = []
            # if there are more than 2 qubits used, treat them all as parallel sq ops
            # one qubit will just be a single sq op
            # 3 or more will be 3 or more parallel sq ops
            if len(shard.qubits_used) != 2:  # noqa: PLR2004
                for qubit in shard.qubits_used:
                    op = qubit.index
                    layer.append(op)
            else:
                for qubit in shard.qubits_used:
                    op.append(qubit.index[0])
                layer.append(op)

            scheduled.add(shard.ID)

        layers.append(layer)

    return layers, shards_in_layer



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from enum import Enum

from .machine import Machine


class QtmMachine(Enum):
    """Available machine architectures."""

    H1_1 = "H1-1"
    H1_2 = "H1-2"


QTM_MACHINES_MAP = {
    QtmMachine.H1_1: Machine(
        size=20,
        tq_options={0, 2, 4, 6, 8, 10, 12, 14, 16, 18},
        # need to get better timing values for below
        # but will have to look them up in hqcompiler
        tq_time=3.0,
        sq_time=1.0,
        qb_swap_time=2.0,
    ),
    QtmMachine.H1_2: Machine(
        size=12,
        tq_options={0, 2, 4, 6, 8, 10},
        # need to get better timing values for below
        # but will have to look them up in hqcompiler
        tq_time=3.0,
        sq_time=1.0,
        qb_swap_time=2.0,
    ),
}



##############################################################################
#
# Copyright (c) 2024 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"

import json

from pytket.circuit import Circuit
from pytket.phir.api import pytket_to_phir


class TestPhirGen:
    def test_classicalexpbox(self) -> None:
        """From https://github.com/CQCL/pytket-phir/issues/86 ."""
        circ = Circuit(1)
        a = circ.add_c_register("a", 2)
        b = circ.add_c_register("b", 2)
        c = circ.add_c_register("c", 3)
        circ.add_classicalexpbox_register(a + b, c.to_list())

        phir = json.loads(pytket_to_phir(circ))
        assert phir["ops"][4] == {
            "cop": "=",
            "returns": ["c"],
            "args": [{"cop": "+", "args": ["a", "b"]}],
        }

    def test_nested_arith(self) -> None:
        """From https://github.com/CQCL/pytket-phir/issues/87 ."""
        circ = Circuit(1)
        a = circ.add_c_register("a", 2)
        b = circ.add_c_register("b", 2)
        c = circ.add_c_register("c", 3)
        circ.add_classicalexpbox_register(a + b // c, c.to_list())

        phir = json.loads(pytket_to_phir(circ))
        assert phir["ops"][4] == {
            "cop": "=",
            "returns": ["c"],
            "args": [{"cop": "+", "args": ["a", {"cop": "/", "args": ["b", "c"]}]}],
        }

    def test_arith_with_int(self) -> None:
        """From https://github.com/CQCL/pytket-phir/issues/88 ."""
        circ = Circuit(1)
        a = circ.add_c_register("a", 2)
        circ.add_classicalexpbox_register(a << 1, a.to_list())

        phir = json.loads(pytket_to_phir(circ))
        assert phir["ops"][2] == {
            "cop": "=",
            "returns": ["a"],
            "args": [{"cop": "<<", "args": ["a", 1]}],
        }

    def test_bitwise_ops(self) -> None:
        """From https://github.com/CQCL/pytket-phir/issues/91 ."""
        circ = Circuit(1)
        a = circ.add_c_register("a", 2)
        b = circ.add_c_register("b", 2)
        c = circ.add_c_register("c", 1)
        expr = a[0] ^ b[0]
        circ.add_classicalexpbox_bit(expr, [c[0]])

        phir = json.loads(pytket_to_phir(circ))
        assert phir["ops"][4] == {
            "cop": "=",
            "returns": [["c", 0]],
            "args": [{"cop": "^", "args": [["a", 0], ["b", 0]]}],
        }



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

# mypy: disable-error-code="misc"
# ruff: noqa: T201

from argparse import ArgumentParser
from importlib.metadata import version

from pecos.engines.hybrid_engine import HybridEngine  # type:ignore [import-not-found]
from pecos.foreign_objects.wasmtime import WasmtimeObj  # type:ignore [import-not-found]

from pytket.qasm.qasm import (
    circuit_from_qasm,
    circuit_from_qasm_wasm,
)

from .api import pytket_to_phir
from .qtm_machine import QtmMachine


def main() -> None:
    """pytket-phir compiler CLI."""
    parser = ArgumentParser(
        prog="phirc",
        description="Emulates QASM program execution via PECOS",
    )
    parser.add_argument(
        "qasm_files", nargs="+", default=None, help="One or more QASM files to emulate"
    )
    parser.add_argument(
        "-w",
        "--wasm-file",
        default=None,
        help="Optional WASM file for use by the QASM programs",
    )
    parser.add_argument(
        "-m",
        "--machine",
        choices=["H1-1", "H1-2"],
        default="H1-1",
        help="Machine name, H1-1 by default",
    )
    parser.add_argument(
        "-o",
        "--tket-opt-level",
        choices=["0", "1", "2"],
        default="0",
        help="TKET optimization level, 0 by default",
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument(
        "--version",
        action="version",
        version=f'{version("pytket-phir")}',
    )
    args = parser.parse_args()

    for file in args.qasm_files:
        print(f"Processing {file}")
        circuit = None
        if args.wasm_file:
            print(f"Including WASM from file {args.wasm_file}")
            circuit = circuit_from_qasm_wasm(file, args.wasm_file)
            wasm_pecos_obj = WasmtimeObj(args.wasm_file)
        else:
            circuit = circuit_from_qasm(file)

        match args.machine:
            case "H1-1":
                machine = QtmMachine.H1_1
            case "H1-2":
                machine = QtmMachine.H1_2

        phir = pytket_to_phir(circuit, machine, int(args.tket_opt_level))
        if args.verbose:
            print("\nPHIR to be simulated:")
            print(phir)

        print("\nPECOS results:")
        print(
            HybridEngine(qsim="state-vector").run(
                program=phir,
                shots=10,
                foreign_object=wasm_pecos_obj if args.wasm_file else None,
            )
        )



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################


class Machine:
    """A machine info class for testing."""

    def __init__(
        self,
        size: int,
        tq_options: set[int],
        tq_time: float,
        sq_time: float,
        qb_swap_time: float,
    ):
        """Create Machine object.

        Args:
            size: number of qubits/slots
            tq_options: options for where to perform tq gates
            tq_time: time for a two qubit gate
            sq_time: time for a single qubit gate
            qb_swap_time: time it takes to swap to qubits
        """
        self.size = size
        self.tq_options = tq_options
        self.sq_options: set[int] = set()
        self.tq_time = tq_time
        self.sq_time = sq_time
        self.qb_swap_time = qb_swap_time

        for i in self.tq_options:
            self.sq_options.add(i)
            self.sq_options.add(i + 1)



##############################################################################
#
# Copyright (c) 2023 Quantinuum LLC All rights reserved.
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
#
##############################################################################

from typing import TYPE_CHECKING

from pytket.extensions.quantinuum.backends.api_wrappers import QuantinuumAPIOffline
from pytket.extensions.quantinuum.backends.quantinuum import (
    QuantinuumBackend,
)
from pytket.passes import DecomposeBoxes

if TYPE_CHECKING:
    from pytket.circuit import Circuit


def rebase_to_qtm_machine(
    circuit: "Circuit", qtm_machine: str, tket_optimization_level: int
) -> "Circuit":
    """Rebases a circuit's gate to the gate set appropriate for the given machine."""
    qapi_offline = QuantinuumAPIOffline()
    backend = QuantinuumBackend(
        device_name=qtm_machine,
        machine_debug=False,
        api_handler=qapi_offline,  # type: ignore [arg-type]
    )

    # Decompose boxes to ensure no problematic phase gates
    DecomposeBoxes().apply(circuit)

    # Optimization level 0 includes rebasing and little else
    # see: https://cqcl.github.io/pytket-quantinuum/api/#default-compilation
    return backend.get_compiled_circuit(circuit, tket_optimization_level)
